Unit AGSuperglobals;

{ Unidad de la clase TSuperGlobalObject, base para instancias de objetos
  "superglobales".  Copyleft 2008 Al Gonzlez.
  
  Web: http://rescatandoadelphi.blogspot.com/

  Esta unidad de cdigo est generosamente comentada para facilitar el
  aprendizaje, pero puede utilizarse libremente en cualquier proyecto. }

Interface

  Type
    {*******************************************************************}
    { Clase TSuperGlobalObject.  Base para clases de objetos cuyas      }
    { instancias pueden ser compartidas entre procesos, a nivel del     }
    { sistema, de forma natural y transparente.  Emplea funciones de    }
    { mapeo de archivos ("file-mapping") de la API de Windows para      }
    { ubicar la instancia del objeto en memoria compartida.  Aprovecha  }
    { la virtualidad del mtodo NewInstance para llevar a cabo tan      }
    { especial asignacin de memoria.                                   }
    {                                                                   }
    { NOTA: Debido a que una instancia de objeto de este tipo es en s  }
    { una vista mapeada ("mapped view") de memoria compartida a nivel   }
    { global del sistema operativo, su uso tiene varias limitaciones:   }
    {   1. Las clases descendientes no deben implementar interfaces.    }
    {   2. Las clases descendientes no deben contener campos que        }
    {      utilicen contadores de referencia, como son las cadenas      }
    {      largas (en lugar de stas utilice ShortString o              }
    {      "String [n]"), los variantes, las interfaces y los arreglos  }
    {      dinmicos, ni tampoco punteros.                              }
    {   3. Solo podrn emplearse mtodos virtuales o referencias        }
    {      (directas o indirectas) a la clase (ClassType) del objeto,   }
    {      cuando todos los procesos que lo compartan sean instancias   }
    {      del mismo mdulo ejecutable.                                 }
    {                                                                   }
    { Debe tenerse en cuenta tambin que los mtodos de un objeto de    }
    { memoria compartida, aunque pueden realizar modificaciones en los  }
    { valores de sus campos, su ejecucin se lleva a cabo en el         }
    { contexto del proceso que los llama.                               }
    {*******************************************************************}
    TSuperGlobalObject = Class
      Public
        { Redefinimos estos dos mtodos virtuales para ubicar a la nueva
          instancia de objeto en la memoria compartida de Windows y
          posteriormente liberar el espacio que ocupa cuando el objeto es
          destruido }

        Class Function NewInstance :TObject; Override;
        Procedure FreeInstance; Override;
    End;

    { Tipo clase de TSuperGlobalObject, auxiliar de la funcin
      CreateSuperGlobalObject }
    TSuperGlobalObjectClass = Class Of TSuperGlobalObject;

  { Funciones y procedimientos }

  { Funcin para crear una instancia de objeto superglobal identificndola
    con un nombre especfico }
  Function CreateSuperGlobalObject (Const Cls :TSuperGlobalObjectClass;
    Const Name :String) :Pointer;

Implementation

  Uses
    Classes, Windows, SysUtils;

  Var
    { Lista de identificadores ("handles") dados por la API de Windows, uno
      por cada objeto de memoria compartida.  No podemos guardar el
      identificador de cada objeto en la instancia misma, porque Windows
      proporciona diferentes identificadores a cada proceso aunque se
      trate del mismo objeto compartido. }
    HandleList :TList;

    { Lo anterior hace que debamos mantener esta lista de todos los objetos
      superglobales, para determinar cul es el identificador de cada
      instancia creada. }
    ObjectList :TList;

    { La funcin CreateSuperGlobalObject guarda en esta variable el nombre
      dado, para despus ser usado por el mtodo NewInstance }
    ObjectName :String;

  { TSuperGlobalObject }

  { Redefinicin del mtodo NewInstance para evitar que el objeto sea
    constituido con un bloque de memoria convencional, haciendo que su
    memoria sea una vista mapeada del archivo de paginacin del sistema
    operativo }
  Class Function TSuperGlobalObject.NewInstance :TObject;
  Var
    AlreadyExists :Boolean;
    Handle :THandle;
    Name :String;
  Begin
    { Si esta instanciacin fue a travs de la funcin
      CreateSuperGlobalObject, tomar el nombre dado, de lo contrario tomar
      el nombre de la clase para el objeto de memoria compartida }
    If ObjectName <> '' Then
      Name := ObjectName
    Else
      Name := ClassName;

    { Crear el (u obtener acceso al) objeto Windows de memoria compartida
      con el tamao de instancia y nombre correspondientes, agregndole a
      ste el sufijo ".mem" para evitar conflictos con otro tipo de objetos
      Windows que podran tener el mismo nombre.  Es prudente darle a cada
      objeto Windows un sufijo segn su tipo (memoria compartida, mutex,
      semforo, etc.) }
    Handle := CreateFileMapping (Invalid_Handle_Value, Nil, Page_ReadWrite,
      0, InstanceSize, Pointer (Name + '.mem'));

    { Despus de llamar a CreateFileMapping, GetLastError devuelve
      error_Already_Exists si el objeto compartido ya haba sido creado por
      este u otro proceso.  Eso no es propiamente un error, pero s un
      indicador muy til. }
    AlreadyExists := GetLastError () = error_Already_Exists;

    If Handle <> 0 Then
    Begin
      { Crear una vista mapeada sobre la totalidad del objeto Windows de
        memoria compartida.  Esta vista ser nuestra "nueva" instancia de
        objeto Delphi superglobal. }
      Result := MapViewOfFile (Handle, file_map_All_Access, 0, 0, 0);

      If Result <> Nil Then
      Begin
        { Si el objeto de memoria compartida no exista previamente,
          llamamos al mtodo InitInstance ms que nada para que ste le
          asigne la clase a la cual pertenece.  No lo llamamos
          incondicionalmente porque InitInstance tambin inicializa en
          ceros los bytes de la instancia dada, cosa que no debemos hacer
          en caso de que el objeto superglobal estuviera ya creado. }
        If Not AlreadyExists Then
          InitInstance (Result);

        { Crear las listas de identificadores y de objetos en caso de que
          an no lo estn }
        If HandleList = Nil Then
        Begin
          HandleList := TList.Create;
          ObjectList := TList.Create;
        End;

        { Aadir a las listas la relacin identificador-objeto }
        HandleList.Add (Pointer (Handle));
        ObjectList.Add (Result);

        Exit;
      End;
    End;

    { Elevar esta excepcin en caso de fallar CreateFileMapping o
      MapViewOfFile }
    Raise Exception.CreateFmt (
      'Fall intento de crear objeto superglobal "%s".', [Name]);
  End;

  { Redefinicin del mtodo FreeInstance para liberar el objeto de memoria
    compartida }
  Procedure TSuperGlobalObject.FreeInstance;
  Var
    I :Integer;
  Begin
    { Liberar la vista mapeada }
    UnmapViewOfFile (Self);

    { Liberar la memoria compartida y eliminar su identificador de la
      lista HandleList }
    I := ObjectList.IndexOf (Self);
    CloseHandle (THandle (HandleList.Extract (HandleList [i])));

    { Eliminar la instancia de la lista ObjectList }
    ObjectList.Delete (I);

    { Destruir las listas HandleList y ObjectList si han quedado vacas }
    If HandleList.Count = 0 Then
    Begin
      FreeAndNil (HandleList);
      FreeAndNil (ObjectList);
    End;
  End;

  { Funciones y procedimientos }

  { Funcin auxiliar para crear instancias de objetos superglobales con
    nombres especficos.  El nombre no puede ser dado como parmetro de un
    constructor porque lo necesita el mtodo NewInstance, el cual se
    ejecuta antes que los constructores. }
  Function CreateSuperGlobalObject (Const Cls :TSuperGlobalObjectClass;
    Const Name :String) :Pointer;
  Begin
    { Guardar en la variable ObjectName el nombre que usar
      TSuperGlobalObject.NewInstance en la siguiente instanciacin }
    ObjectName := Name;

    Try
      { Crear el objeto superglobal }
      Result := Cls.Create;
    Finally
      { Limpiar la variable ObjectName para no afectar a otras
        instanciaciones que se realicen sin la ayuda de esta funcin }
      ObjectName := '';
    End;
  End;

End.